from commons import update_file
from config import renames


def renamer(match, result_file=None):
    # ^( *)(' + '|'.join(renames.keys()) + r')( *: *)([^;\n]+);( *//.*)?$
    original = match.group(0)
    spaces1 = match.group(1)
    typ = match.group(2)
    spaces2 = match.group(3)
    values = match.group(4)
    comment = match.group(5) or ''

    try:
        replaced = f"{spaces1}{renames[typ]}{spaces2}{values};{comment}"
        update_file(result_file, original, replaced)
        return replaced
    except KeyError:
        return original



sides = {
    '-top': '-block-start',
    '-right': '-inline-end',
    '-bottom': '-block-end',
    '-left': '-inline-start',
}

renames = {
    # These are mostly unimplemented, might require https://github.com/csstools/postcss-logical
    'left': 'inset-inline-start',
    'right': 'inset-inline-end',
    'top': 'inset-block-start',
    'bottom': 'inset-block-end',

    'min-height': 'min-block-size',
    'max-height': 'max-block-size',
    'min-width': 'min-inline-size',
    'max-width': 'max-inline-size',
}

inline_start_end = {
    'left': 'inline-start',
    'right': 'inline-end',
}

aligns = {
    'text-align': {
        'left': 'start',
        'right': 'end',
    },
    # These are mostly unimplemented, might require https://github.com/csstools/postcss-logical
    'float': inline_start_end,
    'clear': inline_start_end,
}

replacer_reg = r'^( *)(margin|padding|border)(-(?:left|right|top|bottom))?(-(?:size|style|color))?( *: *)([^;\n]+);( *//.*)?$'
renamer_reg = r'^( *)(' + '|'.join(renames.keys()) + r')( *: *)([^;\n]+);( *//.*)?$'
aligner_reg = r'^( *)(text-align|float|clear)( *: *)(left|right)( *);'
src_files_pattern = ('*.scss', '*.css')



def collapse_if_equal(val1, val2):
    return val1 if val1 == val2 else f"{val1} {val2}"


def update_file(result_file, original, replaced):
    if not result_file or original == replaced:
        return

    result = f"<tr><td> {original.lstrip()} </td><td> {replaced.lstrip()} </td></tr>\n"
    result_file.write(result)



def tokenize_values(string):
    # Original idea from https://stackoverflow.com/a/38212061/177275
    # Assume correct syntax / matching brackets
    brackets = 0
    start = 0
    results = []

    for idx, char in enumerate(string):
        if char == ' ':
            if start is not None and brackets == 0:
                results.append(string[start:idx])
                start = None
        else:
            if start is None:
                start = idx
            if char == '(' or char == '{':
                brackets += 1
            elif char == ')' or char == '}':
                brackets -= 1
                if brackets < 0:
                    raise ValueError(f'failed to tokenize "{string}"')
    if start is not None:
        results.append(string[start:])
    results2 = []
    idx = 0
    while idx < len(results):
        if len(results[idx]) > 1 or results[idx] not in ('-', '+', '/', '*'):
            results2.append(results[idx])
        else:
            results2[-1] += ' ' + results[idx] + ' ' + results[idx + 1]
            idx += 1
        idx += 1
    return results2



from process_files import analyze_files


def analyze(root_path, is_recursive, result_file):
    f = open(result_file, "a+")
    f.truncate(0)

    header = "<table><thead>\n"
    header += "<tr><th> physical </th><th> logical </th></tr>\n"
    header += "</thead><tbody>\n"
    f.write(header)

    analyze_files(root_path, is_recursive, f)

    footer = "</tbody></table>\n"
    f.write(footer)

    f.close()

    with open(result_file, 'r') as file:
        lines = file.read().split("\n")

        for i in range(1, len(lines) + 1):
            text_match = "EMPTY_FILE"

            if lines[-i] == text_match:
                lines[-(i + 1)] = ""
                lines[-(i + 0)] = ""

        while "" in lines:
            lines.remove("")

        code_updated = "\n".join(lines)

        with open(result_file, 'w') as out:
            out.write(code_updated)
            out.close()



import argparse
from pathlib import Path
from analyze import analyze
from process_files import update_files

parser = argparse.ArgumentParser(description='Convert CSS physical properties to logical')
parser.add_argument('source',
                    nargs='*',
                    default='.',
                    help='Path to source directory / file. Default "."')
parser.add_argument('-r', '--recursive',
                    default='True',
                    action='store_true',
                    help='Process all files from source directory. Default: True')
parser.add_argument('-a', '--analyze',
                    action='store_true',
                    default='True',
                    help='Dry Run - Create a report with all changes (without actual source file changes). '
                         ' Default: True')
parser.add_argument('-f', '--filename',
                    default='report.html',
                    help='Path to file where to save reports (used only with -a/--analyze flag). Default: "report.html"')

args = parser.parse_args()

root_path = Path(args.source or ".")
result_file = args.filename or "report.html"
is_recursive = args.recursive
is_analyze = args.analyze

if is_analyze:
    analyze(root_path, is_recursive, result_file)

else:
    update_files(root_path, is_recursive)



import re
from os.path import relpath

from aligner import aligner
from config import replacer_reg, renamer_reg, aligner_reg, src_files_pattern
from renamer import renamer
from replacer import replacer


def process_file(source_file, result_file):
    def re_callback(callback):
        if result_file:
            return lambda match: callback(match, result_file)
        return callback

    code = source_file.read_text()
    res = re.sub(replacer_reg, re_callback(replacer), code, flags=re.MULTILINE)
    res = re.sub(renamer_reg, re_callback(renamer), res, flags=re.MULTILINE)
    res = re.sub(aligner_reg, re_callback(aligner), res, flags=re.MULTILINE)

    if res != code:
        return res

    return ""


def process_results(file_path, source_file, result_file):
    header = f"<tr style='background: silver'><td> File: </td><td> {file_path} </td></tr>\n"
    result_file.write(header)

    changed = process_file(source_file, result_file)

    if not changed:
        result_file.write("EMPTY_FILE\n")


def process_files(root_path, is_recursive, result_file):
    print(f"START processing files in '{root_path}'")
    if root_path.is_file():
        process_file(root_path, result_file)
    else:
        for pattern in src_files_pattern:
            all_files_gen = root_path.rglob(pattern) if is_recursive else root_path.glob(pattern)
            all_files_list = list(all_files_gen)
            print(f"Found {len(all_files_list)} files for pattern {pattern}")
            for file in all_files_list:
                if file.is_file():
                    source_file = root_path / file
                    rel_path = relpath(file, root_path)

                    if result_file:
                        process_results(rel_path, source_file, result_file)
                    else:
                        changed = process_file(source_file, None)
                        if changed:
                            print(f"Modifying {rel_path}")
                            source_file.write_text(changed)
    print("DONE processing files.")


def analyze_files(root_path, is_recursive, result_file):
    print("Analyzing files")
    return process_files(root_path, is_recursive, result_file)


def update_files(root_path, is_recursive):
    print("Updating files")
    return process_files(root_path, is_recursive, None)



from commons import update_file
from config import sides
from tokenize_values import tokenize_values


def replacer(match, result_file=None):
    # ^( *)(margin|padding|border)(-(?:left|right|top|bottom))?(-(?:size|style|color))?( *: *)([^;\n]+);( *//.*)?$
    original = match.group(0)
    replaced = original
    spaces1 = match.group(1)
    typ = match.group(2)
    side = match.group(3) or ''
    extra = match.group(4) or ''
    spaces2 = match.group(5)
    values = match.group(6)
    comment = match.group(7) or ''
    # special lint controlling comment should be repeated
    dup_comment = comment if ('sass-lint:' in comment) else ''
    nl = "<br/>" if result_file else ""

    if not side:
        if typ == 'border':
            replaced = original
        else:
            tokens = tokenize_values(values)
            important = ' ' + tokens[-1] if tokens[-1] == '!important' else ''

            if important:
                tokens = tokens[:-1]

            if len(tokens) == 1:
                replaced = original  # single token stays as is
            elif len(tokens) == 2:  # top-bottom  right-left
                # The *-block shorthand is not yet supported,  see
                #   https://developer.mozilla.org/en-US/docs/Web/CSS/margin-block
                replaced = f"{spaces1}{typ}-block-start{extra}{spaces2}{tokens[0]}{important};{comment}{nl}\n" + \
                           f"{spaces1}{typ}-block-end{extra}{spaces2}{tokens[0]}{important};{dup_comment}{nl}\n" + \
                           f"{spaces1}{typ}-inline-start{extra}{spaces2}{tokens[1]}{important};{dup_comment}{nl}\n" + \
                           f"{spaces1}{typ}-inline-end{extra}{spaces2}{tokens[1]}{important};{dup_comment}{nl}"
            elif len(tokens) == 3:  # top  left-right  bottom
                replaced = f"{spaces1}{typ}-block-start{extra}{spaces2}{tokens[0]}{important};{comment}{nl}\n" + \
                           f"{spaces1}{typ}-block-end{extra}{spaces2}{tokens[2]}{important};{dup_comment}{nl}\n" + \
                           f"{spaces1}{typ}-inline-start{extra}{spaces2}{tokens[1]}{important};{dup_comment}{nl}\n" + \
                           f"{spaces1}{typ}-inline-end{extra}{spaces2}{tokens[1]}{important};{dup_comment}{nl}"
            elif len(tokens) == 4:  # top  left-right  bottom
                replaced = f"{spaces1}{typ}-block-start{extra}{spaces2}{tokens[0]}{important};{dup_comment}{comment}{nl}\n" + \
                           f"{spaces1}{typ}-block-end{extra}{spaces2}{tokens[2]}{important};{dup_comment}{nl}\n" + \
                           f"{spaces1}{typ}-inline-start{extra}{spaces2}{tokens[3]}{important};{dup_comment}{nl}\n" + \
                           f"{spaces1}{typ}-inline-end{extra}{spaces2}{tokens[1]}{important};{dup_comment}{nl}"
            else:
                raise ValueError(f'Unexpected number of tokens {len(tokens)} in {original} -- {tokens}')

    else:
        replaced = f"{spaces1}{typ}{sides[side]}{extra}{spaces2}{values};{comment}"

    update_file(result_file, original, replaced)
    return replaced



from commons import update_file
from config import aligns


def aligner(match, result_file=None):
    # ^( *)(text-align|float|clear)( *: *)(left|right)( *);
    original = match.group(0)
    spaces1 = match.group(1)
    typ = match.group(2)
    spaces2 = match.group(3)
    values = match.group(4)
    spaces3 = match.group(5)

    try:
        replaced = f"{spaces1}{typ}{spaces2}{aligns[typ][values]}{spaces3};"
        update_file(result_file, original, replaced)
        return replaced
    except KeyError:
        return original
